#include <DCC_Decoder.h>

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Defines and structures
//
#define kDCC_INTERRUPT            0
#define VERSION                   "1.05"
#define FIRST_DECODING_ERR_CODE   kDCC_ERR_DETECTION_FAILED
#define LAST_DECODING_ERR_CODE    kDCC_ERR_MISSING_END_BIT
// Error type total number : From 100 to 106 => so 106-100+1=7
#define DECODING_ERR_NUMBER       LAST_DECODING_ERR_CODE - FIRST_DECODING_ERR_CODE + 1

typedef struct
{
    int count;
    byte validBytes;
    byte data[6];
} DCCPacket;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// The dcc decoder object and global data
//
int gPacketCount = 0;
int gIdlePacketCount = 0;
int gLongestPreamble = 0;
int gPinLedInterruptMonitor = 13;
volatile int gStateLedInterruptMonitor = LOW;

// Array of DCC packets
DCCPacket gPackets[75];
// Array of decoding error (error type (= index) and counter associated)
int gDecodingErrors[DECODING_ERR_NUMBER];
// gDecodingErrors[0] = kDCC_ERR_DETECTION_FAILED error counter
// gDecodingErrors[1] = kDCC_ERR_BASELINE_ADDR error counter
// Etc...

static unsigned long lastMillis = millis();
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Packet handlers
//

// ALL packets are sent to the RawPacket handler. Returning true indicates that packet was handled. DCC library starts watching for 
// next preamble. Returning false and library continue parsing packet and finds another handler to call.
boolean RawPacket_Handler(byte byteCount, byte* packetBytes)
{
        // Bump global packet count
    ++gPacketCount;
    
    int thisPreamble = DCC.LastPreambleBitCount();
    if( thisPreamble > gLongestPreamble )
    {
        gLongestPreamble = thisPreamble;
    }
    
        // Walk table and look for a matching packet
    for( int i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
    {
        if( gPackets[i].validBytes )
        {
                // Not an empty slot. Does this slot match this packet? If so, bump count.
            if( gPackets[i].validBytes==byteCount )
            {
                char isPacket = true;
                for( int j=0; j<byteCount; j++)
                {
                    if( gPackets[i].data[j] != packetBytes[j] )
                    {
                        isPacket = false;
                        break;
                    } 
                }
                if( isPacket )
                {
                   gPackets[i].count++;
                   return false;
                }
            }
        }else{
                // Empty slot, just copy over data
            gPackets[i].count++;
            gPackets[i].validBytes = byteCount;
            for( int j=0; j<byteCount; j++)
            {
                gPackets[i].data[j] = packetBytes[j];
            }
            return false;
        }
    }    
    
    return false;
}

// Idle packets are sent here (unless handled in rawpacket handler). 
void IdlePacket_Handler(byte byteCount, byte* packetBytes)
{
    ++gIdlePacketCount;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Other information handlers
//

// When the device received a defined quantity of interrupt
// it will call this method => we use it to blink a LED
void InterruptMonitor_Handler()
{
    gStateLedInterruptMonitor = !gStateLedInterruptMonitor;
    digitalWrite(gPinLedInterruptMonitor, gStateLedInterruptMonitor);
}

// This method is called when DCC lib does a reset
void DecodingEngineCompletion_Handler(byte resultOfLastPacket)
{
    // If it's an OK reset, just ignore it
    if (resultOfLastPacket < FIRST_DECODING_ERR_CODE)
        return;

    // Array index is simply the ERR_CODE - FIRST_ERR_CODE       
    int iIndex = (int)resultOfLastPacket - FIRST_DECODING_ERR_CODE;
     
    if ((iIndex >= 0) && (iIndex < sizeof(gDecodingErrors)))
        gDecodingErrors[iIndex]++; 
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Setup
//
void setup() 
{ 
   Serial.begin(115200);
   
   DCC.SetRawPacketHandler(RawPacket_Handler);   
   DCC.SetIdlePacketHandler(IdlePacket_Handler);
   DCC.SetDecodingEngineCompletionStatusHandler(DecodingEngineCompletion_Handler);

   DCC.SetupMonitor( kDCC_INTERRUPT );

   // Used to blink LED when receiving packet
   DCC.SetInterruptMonitor(InterruptMonitor_Handler);
   pinMode(gPinLedInterruptMonitor, OUTPUT);
      
   // Initialize the error counter array
   for (int i = 0; i < DECODING_ERR_NUMBER; i++)
       gDecodingErrors[i] = 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Called when new data available on RX. 
// We use it to answer "?" question from computer
void serialEvent()
{
    // Only read ONE char !
    char inChar = (char)Serial.read();  
    
    if (inChar == '?')
    {
        Serial.println("[ArduinoVersionBegin]");
        Serial.print("DCC Monitor version : ");
        Serial.println(VERSION);
        Serial.print("DCC Decoder library version : ");
        Serial.println(DCC_DECODER_VERSION);
        Serial.println("[ArduinoVersionEnd]");
    }
}

void DumpAndResetTable()
{
    char buffer60Bytes[60];
    
    /* General informations */
    //----------------------//
     
    Serial.print("Longest Preamble:  ");
    Serial.println(gLongestPreamble, DEC);
    
    /* Erorr list */
    //------------//    
        
    Serial.println("[ErrorListBegin]");       
    for (int i=0; i<DECODING_ERR_NUMBER; i++)
    {
        Serial.print(i + FIRST_DECODING_ERR_CODE, DEC);
        Serial.print(" ");  
        Serial.println(gDecodingErrors[i], DEC);      
      
        // Reset array
        gDecodingErrors[i] = 0;  
    }                  
    Serial.println("[ErrorListEnd]");
   
    /* DCC Packets list */
    //------------------//   
   
    Serial.println("[PacketListBegin]");
    for( int i=0; i<(int)(sizeof(gPackets)/sizeof(gPackets[0])); ++i )
    {
        if( gPackets[i].validBytes > 0 )
        {
            Serial.print(gPackets[i].count, DEC);
            if( gPackets[i].count < 10 )
            {
                Serial.print("        ");
            }else{
                if( gPackets[i].count < 100 )
                {
                    Serial.print("       ");
                }else{
                    Serial.print("      ");
                }
            }
            Serial.println( DCC.MakePacketString(buffer60Bytes, gPackets[i].validBytes, &gPackets[i].data[0]) );
        }

        gPackets[i].validBytes = 0;
        gPackets[i].count = 0;
    }    
    Serial.println("[PacketListEnd]");
    Serial.println("");
    
    /* Resets */
    //--------//      
    
    gPacketCount = 0;
    gIdlePacketCount = 0;
    gLongestPreamble = 0;    
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Main loop
//
void loop()
{
    DCC.loop();
    
    if( millis()-lastMillis > 2000 )
    {
        DumpAndResetTable();
        lastMillis = millis();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
